/****************************************************************************
  FileName     [ myHash.h ]
  PackageName  [ util ]
  Synopsis     [ Define Hash and Cache ADT ]
  Author       [ Chung-Yang (Ric) Huang ]
  Copyright    [ Copyleft(c) 2009-2012 LaDs(III), GIEE, NTU, Taiwan ]
****************************************************************************/

#ifndef MY_HASH_H
#define MY_HASH_H

#include <vector>

using namespace std;

//--------------------
// Define Hash classes
//--------------------
// To use Hash ADT, you should define your own HashKey class.
// It should at least overload the "()" and "==" operators.
//
// class HashKey
// {
// public:
//    HashKey() {}
// 
//    size_t operator() () const { return 0; }
// 
//    bool operator == (const HashKey& k) { return true; }
// 
// private:
// };
//
class HashKey
{
public:
	HashKey(And* g = 0) : _f1((*g)._fanin1), _f2((*g)._fanin2) {}
	CirGate* getdata1() const { return _f1; }
	CirGate* getdata2() const { return _f2; }
	size_t operator() () const { return (*(size_t*)(_f1)) + (*(size_t*)(_f2)); }
	bool operator == (const HashKey& k)
	{
		if(_f1==k.getdata1() && _f2==k.getdata2()) return true;
		if(_f1==k.getdata2() && _f2==k.getdata1()) return true;
		return false;
	}

private:
	CirGate* _f1;
	CirGate* _f2;
};

class SimHashKey
{
public:
	SimHashKey(CirGate* g = 0) : _gate(g) {}
	unsigned getvalue() const { return _gate->_sim; }
	size_t operator() () const
	{
		if(_gate->_sim <= ~(_gate->_sim)) return _gate->_sim;
		else return ~(_gate->_sim);
	}
	bool operator == (const SimHashKey& k)
	{
		if(k.getvalue() == _gate->_sim) return true;
		else if(~(k.getvalue()) == _gate->_sim) return true;
		else return false;
	}

private:
	CirGate* _gate;
};

template <class HashKey, class HashData>
class Hash
{
typedef pair<HashKey, HashData> HashNode;

public:
   Hash() : _numBuckets(0), _buckets(0) {}
   Hash(size_t b) : _numBuckets(0), _buckets(0) { init(b); }
   ~Hash() { reset(); }


   // TODO: implement the Hash<HashKey, HashData>::iterator
   // o An iterator should be able to go through all the valid HashNodes
   //   in the Hash
   // o Functions to be implemented:
   //   - constructor(s), destructor
   //   - operator '*': return the HashNode
   //   - ++/--iterator, iterator++/--
   //   - operators '=', '==', !="
   //
   // (_bId, _bnId) range from (0, 0) to (_numBuckets, 0)
   //
   class iterator
   {
      friend class Hash<HashKey, HashData>;

   public:
		iterator(Hash* h = 0, HashNode hn = 0, size_t a = 0, size_t b = 0)
			:_h(h), _hashnode(hn), _bnum(a), _nodenum(b) {}
		iterator(const iterator& i)
			: _h(i._h), _hashnode(i._hashnode), _bnum(i._bnum), _nodenum(i._nodenum) {}
		~iterator() {}

		const HashData& operator * () const { return _hashnode.second; }
		HashData& operator * () { return _hashnode.second; }
		iterator& operator ++ () { return 0; }
		iterator operator ++ (int) { return 0; }
		iterator& operator --() { return 0; }
		iterator operator -- (int) { return 0; }
		iterator& operator = (const iterator& i) { return 0; }
		bool operator != (const iterator& i) const { return 0; }
		bool operator == (const iterator& i) const { return 0; }

   private:
		Hash*    _h;
		HashNode _hashnode;
		size_t   _bnum;
		size_t   _nodenum;
   };

   // TODO: implement these functions
   //
   // Point to the first valid data
   iterator begin() const { return iterator(); }
   // Pass the end
   iterator end() const { return iterator(); }
   // return true if no valid data
   bool empty() const { return size()==0; }
   // number of valid data
   size_t size() const
	{
		size_t s = 0;
		for(size_t i=0; i<_numBuckets; ++i)
			s += _buckets[i].size();
		return s;
	}
   size_t numBuckets() const { return _numBuckets; }

   vector<HashNode>& operator [] (size_t i) { return _buckets[i]; }
   const vector<HashNode>& operator [](size_t i) const { return _buckets[i]; }

   void init(size_t b)
	{
		_buckets = new vector<HashNode>[b];
		_numBuckets = b;
	}
   void reset() { delete[] _buckets; }

   // check if k is in the hash...
   // if yes, update n and return true;
   // else return false;
   bool check(const HashKey& k, HashData& n)
	{
		for(size_t i=0; i<_buckets[bucketNum(k)].size(); ++i)
			if(_buckets[bucketNum(k)][i].first == k)
			{
				n = _buckets[bucketNum(k)][i].second;
				return true;
			}
		return false;
	}

   // return true if inserted successfully (i.e. k is not in the hash)
   // return false is k is already in the hash ==> will not insert
   bool insert(const HashKey& k, const HashData& d)
	{
		for(size_t i=0; i<_buckets[bucketNum(k)]; ++i)
			if(_buckets[bucketNum(k)][i].first == k) return false;
		forceInsert();
		return true;
	}

   // return true if inserted successfully (i.e. k is not in the hash)
   // return false is k is already in the hash ==> still do the insertion
   bool replaceInsert(const HashKey& k, const HashData& d)
	{
		for(size_t i=0; i<_buckets[bucketNum(k)]; ++i)
			if(_buckets[bucketNum(k)][i].first == k)
			{
				_buckets[bucketNum(k)][i].second = d;
				return false;
			}
		forceInsert();
		return true;
	}

   // Need to be sure that k is not in the hash
   void forceInsert(const HashKey& k, const HashData& d)
	{
		_buckets[bucketNum(k)].push_back(HashNode(k, d));
	}

private:
   // Do not add any extra data member
   size_t                   _numBuckets;
   vector<HashNode>*        _buckets;

   size_t bucketNum(const HashKey& k) const {
      return (k() % _numBuckets); }

};


//---------------------
// Define Cache classes
//---------------------
// To use Cache ADT, you should define your own HashKey class.
// It should at least overload the "()" and "==" operators.
//
// class CacheKey
// {
//    public:
//    CacheKey() {}
//    
//    size_t operator() () const { return 0; }
//   
//    bool operator == (const CacheKey&) const { return true; }
//       
// private:
// }; 
// 
template <class CacheKey, class CacheData>
class Cache
{
typedef pair<CacheKey, CacheData> CacheNode;

public:
   Cache() : _size(0), _cache(0) {}
   Cache(size_t s) : _size(0), _cache(0) { init(s); }
   ~Cache() { reset(); }

   // NO NEED to implement Cache::iterator class

   // TODO: implement these functions
   //
   // Initialize _cache with size s
   void init(size_t s)
	{
		_size = s;
		_cache = new CacheNode[s];
		for(size_t i=0; i<s; ++i)
			_cache[i] = 0;
	}
   void reset() { _size = 0; delete _cache; }

   size_t size() const { return _size; }

   CacheNode& operator [] (size_t i) { return _cache[i]; }
   const CacheNode& operator [](size_t i) const { return _cache[i]; }

   // return false if cache miss
   bool read(const CacheKey& k, CacheData& d) const
	{
		if(_cache[k()%_size] == 0) return false;
		else
		{
			d = _cache[k()%_size].second;
			return true;
		}
	}
   // If k is already in the Cache, overwrite the CacheData
   void write(const CacheKey& k, const CacheData& d) { _cache[k()%_size] = CacheNode(k, d); }

private:
   // Do not add any extra data member
   size_t         _size;
   CacheNode*     _cache;
};


#endif // MY_HASH_H
