/****************************************************************************
  FileName     [ cirGate.h ]
  PackageName  [ cir ]
  Synopsis     [ Define basic gate data structures ]
  Author       [ Chung-Yang (Ric) Huang ]
  Copyright    [ Copyleft(c) 2008-2012 LaDs(III), GIEE, NTU, Taiwan ]
****************************************************************************/

#ifndef CIR_GATE_H
#define CIR_GATE_H

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include "cirDef.h"
#include "sat.h"

using namespace std;

class CirGate;

//------------------------------------------------------------------------
//   Define classes
//------------------------------------------------------------------------
class CirGate
{
	friend class SimHashKey;

public:
   CirGate(int i=0, int j=0) : ID(i), _line(j)
	{
		_sim = 0;
		_ref = 0;
		_grpptr = 0;
	}
   virtual ~CirGate() {}

   // Basic access methods
   virtual string getTypeStr() const = 0;
   unsigned getLineNo() const { return _line; }
	int getID() const { return ID; }
	unsigned getsim() const { return _sim; }
	void setgrpptr(vector<CirGate*>* g) { _grpptr = g;}
	vector<CirGate*>* getgrpptr() const { return _grpptr; }
	Var getvar() const { return _var; }
	void setvar(const Var& v) { _var = v; }
	virtual string getsym() { return 0; }
	virtual bool getfinID(int& id) { return 0; }
	virtual bool getfinID2(int& id) { return 0; }

   // Printing functions
   virtual void printGate() const = 0;
   void reportGate() const;
   virtual void reportFanin(int level, int count=0, bool inv=0) const = 0;
   void reportFanout(int level, int count=0, bool inv=0) const;
	
	//virtual func for setting
	virtual void setFanin(CirGate*& fin) {}
	virtual void setFanin(CirGate*& fin1, CirGate*& fin2) {}
	virtual void setFinInv() {}
	virtual void setFin1Inv() {}
	virtual void setFin2Inv() {}
	virtual void setsym(string& str) {}

	void setFanout(CirGate* fout);
	static void setGlobal() { ++_globalRef; }
	void setmark() const { _ref = _globalRef; }
	bool ismarked() const { return (_ref == _globalRef); }

	//virtual func for checking floating
	virtual bool isfinfl() { return 0; } 
	virtual bool isfin1fl() const { return 0; }
	virtual bool isfin2fl() const { return 0; }
	virtual bool isfoutfl() { return 0; }

	//check inv
	virtual bool isfinInv() { return 0; }
	virtual bool isfin1Inv() { return 0; }
	virtual bool isfin2Inv() { return 0; }
	
	virtual bool whichfin(CirGate*) { return 0; }
	
	//sweep
	virtual void sweepaig() {}
	void erasefout(CirGate* fout);
	virtual void erasefin(CirGate* fin) {}

	//optimize
	virtual void optgate(CirGate*&, bool&) {}
	virtual void changefin(CirGate*& fin) {}
	virtual void changefin1(CirGate*& fin1) {}
	virtual void changefin2(CirGate*& fin2) {}

	//simulate
	virtual void simulate() {}
	virtual void simulate(const unsigned&) {}

	virtual void DFS(vector<CirGate*>& _list) = 0;

	static unsigned _globalRef;
private:

protected:
	int ID;
	int _line;
	Var _var;
	unsigned _sim;
	vector<CirGate*>* _grpptr;
	mutable unsigned _ref;
	vector<CirGate*> _fanout;
	
	bool isInv(const CirGate* const & cirg) const
	{
		size_t temp = size_t(cirg);
		return (temp&1)==1;
	}
	CirGate* isInvRe(const CirGate* const & cirg) const
	{
		size_t temp = size_t(cirg);
		return (CirGate*)(temp&=(~1));
	}
};

class CompGate
{
public:
	bool operator() (const CirGate* a, const CirGate* b)
	{ return (a->getID() < b->getID()); }
};

class Pi : public CirGate
{
public:
	Pi(int i=0, int j=0) : CirGate(i, j), _fanin(0), _sym("") {}
	virtual ~Pi() {}
//get data
   string getTypeStr() const { return string("PI"); }
	string getsym() { return _sym; }
//print
	void printGate() const
	{
		string temp = "= ";
		string a = static_cast<ostringstream*>(&(ostringstream()<<ID))->str();
		if(_sym != "")
		{
			temp.append("PI(").append(a).append(")");
			temp.append("\"").append(_sym).append("\"");
			a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
			temp.append(", line ").append(a);
			cout<<setw(49)<<left<<temp;
			cout<<right<<"="<<endl;
		}
		else
		{
			temp.append("PI(").append(a).append(")");
			a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
			temp.append(", line ").append(a);
			cout<<setw(49)<<left<<temp;
			cout<<right<<"="<<endl;
		}
		cout<<"= FECs:                                          ="<<endl;
	}
	void reportFanin(int level, int count=0, bool inv=0) const
	{
		assert(level >= 0);
		for(int i=0; i<2*count; ++i) cout<<" ";
		if(inv) cout<<"!";
		cout<<"PI "<<ID<<endl;
	}
//set
	void setsym(string& str) { _sym = str; }
//bool
	virtual bool isfoutfl() 
	{
		if(_fanout.size()==0) return true;
		return false;
	}
//simulate
	virtual void simulate(const unsigned& i) { _sim = i; }
//DFS
	virtual void DFS(vector<CirGate*>& _list)
	{
		setmark();
		_list.push_back(this);
	}

private:
	CirGate* _fanin;
	string _sym;
};

class Po : public CirGate
{
public:
	Po(int i=0, int j=0) : CirGate(i, j), _sym("") {}
	virtual ~Po() {}
//get data
   string getTypeStr() const { return string("PO"); }
	string getsym() { return _sym; }
	bool getfinID(int& id)
	{
		if(isInv(_fanin))
		{
			id = isInvRe(_fanin)->getID();
			return 0;
		}
		id = _fanin->getID();
		return 1;
	}
//print
	void printGate() const
	{
		string temp = "= ";
		string a = static_cast<ostringstream*>(&(ostringstream()<<ID))->str();
		if(_sym != "")
		{
			temp.append("PO(").append(a).append(")");
			temp.append("\"").append(_sym).append("\"");
			a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
			temp.append(", line ").append(a);
			cout<<setw(49)<<left<<temp;
			cout<<right<<"="<<endl;
		}
		else
		{
			temp.append("PO(").append(a).append(")");
			a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
			temp.append(", line ").append(a);
			cout<<setw(49)<<left<<temp;
			cout<<right<<"="<<endl;
		}
		cout<<"= FECs:                                          ="<<endl;
	}
	void reportFanin(int level, int count=0, bool inv=0) const
	{
		assert(level >= 0);
		if(count == 0) setGlobal();
		for(int i=0; i<2*count; ++i) cout<<" ";
		cout<<"PO "<<ID;
		if(level > 0 && ismarked()) cout<<" (*)";
		cout<<endl;
		if(level > 0 && !ismarked())
		{
			setmark();
			if(isInv(_fanin)) inv = 1;
			else inv = 0;
			isInvRe(_fanin)->reportFanin(level-1, count+1, inv);
		}
	}
//set	
	void setsym(string& str) { _sym = str; }
	
	virtual void setFanin(CirGate*& fin)
	{
		_fanin = fin;
		if(fin!=0)
			fin->setFanout(this);
	}
	virtual void setFinInv()
	{
		size_t temp = size_t(_fanin);
		temp |= 1;
		_fanin = (CirGate*)(temp);
	}
//bool
	virtual bool isfinfl()
	{
		if(isInvRe(_fanin)->getTypeStr() == "UNDEF") return true;
		//if(_fanin->isfinfl()) return true;
		return false;
	}
	virtual bool isfinInv() { return isInv(_fanin); }
//sweep
	virtual void erasefin(CirGate* fin)
	{
		_fanin = 0;
	}
//optimize
	virtual CirGate* optcheck() { return 0; }
	virtual void changefin(CirGate*& fin) { _fanin = fin; }
//simulate
	virtual void simulate()
	{
		if(isfinInv()) { _sim = ~(isInvRe(_fanin)->getsim()); }
		else { _sim = _fanin->getsim(); }
	}
//DFS
	virtual void DFS(vector<CirGate*>& _list)
	{
		if(!isfinfl())
		{
			if(!isInvRe(_fanin)->ismarked())
				isInvRe(_fanin)->DFS(_list);
		}
		setmark();
		_list.push_back(this);
	}

private:
	CirGate* _fanin;
	string _sym;
};

class And : public CirGate
{
	friend class HashKey;

public:
	And(int i=0, int j=0) : CirGate(i, j) {}
	virtual ~And() {}
//get data
   string getTypeStr() const { return string("AIG"); }
	bool getfinID(int& id)
	{
		if(isInv(_fanin1))
		{
			id = isInvRe(_fanin1)->getID();
			return 0;
		}
		id = _fanin1->getID();
		return 1;
	}
	bool getfinID2(int& id)
	{
		if(isInv(_fanin2))
		{
			id = isInvRe(_fanin2)->getID();
			return 0;
		}
		id = _fanin2->getID();
		return 1;
	}
//print
	void printGate() const
	{
		string temp = "= ";
		string a = static_cast<ostringstream*>(&(ostringstream()<<ID))->str();
		temp.append("AIG(").append(a).append(")");
		a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
		temp.append(", line ").append(a);
		cout<<setw(49)<<left<<temp;
		cout<<right<<"="<<endl;
		cout<<"= FECs: ";
		if(_grpptr == 0)
			cout<<"                                         ="<<endl;
		else
		{
			vector<CirGate*> list = *_grpptr;
			sort(list.begin(), list.end(), CompGate());
			temp = "";
			for(size_t i=0; i<list.size(); ++i)
			{
				a = static_cast<ostringstream*>(&(ostringstream()<<(list[i]->getID())))->str();
				if(list[i]->getsim() == _sim && list[i]->getID() != ID)
					temp.append(a).append(" ");
				else if(list[i]->getID() != ID)
					temp.append("!").append(a).append(" ");
			}
			cout<<setw(41)<<left<<temp;
			cout<<right<<"="<<endl;
		}
	}
	void reportFanin(int level, int count=0, bool inv=0) const
	{
		assert(level >= 0);
		if(count == 0) setGlobal();
		for(int i=0; i<2*count; ++i) cout<<" ";
		if(inv) cout<<"!";
		cout<<"AIG "<<ID;
		if(level > 0 && ismarked()) cout<<" (*)";
		cout<<endl;
		if(level > 0 && !ismarked())
		{
			setmark();
			
			inv = isInv(_fanin1);
			isInvRe(_fanin1)->reportFanin(level-1, count+1, inv);
			
			inv = isInv(_fanin2);
			isInvRe(_fanin2)->reportFanin(level-1, count+1, inv);
		}
	}
//set
	virtual void setFanin(CirGate*& fin1, CirGate*& fin2)
	{
		_fanin1 = fin1;
		_fanin2 = fin2;
		if(fin1!=0)
			fin1->setFanout(this);
		if(fin2!=0)
			fin2->setFanout(this);
	}
	virtual void setFin1Inv()
	{
		size_t temp = size_t(_fanin1);
		temp |= 1;
		_fanin1 = (CirGate*)(temp);
	}	
	virtual void setFin2Inv()
	{
		size_t temp = size_t(_fanin2);
		temp |= 1;
		_fanin2 = (CirGate*)(temp);
	}	
//bool
	virtual bool isfinfl()
	{
		if(isfin1fl() || isfin2fl()) return true;
		/*if(isInvRe(_fanin1)->getTypeStr()!="PI")
			if(isInvRe(_fanin1)->isfinfl()) return true;
		if(isInvRe(_fanin2)->getTypeStr()!="PI")
			if(isInvRe(_fanin2)->isfinfl()) return true;*/
		return false;
	}
	virtual bool isfin1fl() const
	{
		if(_fanin1 == 0) return true;
		if(isInvRe(_fanin1)->getTypeStr() == "UNDEF") return true;
		return false;
	}
	virtual bool isfin2fl() const
	{
		if(_fanin2 == 0) return true;
		if(isInvRe(_fanin2)->getTypeStr() == "UNDEF") return true;
		return false;
	}
	virtual bool isfoutfl()
	{
		if(_fanout.size()==0) return true;
		/*for(size_t i=0; i<_fanout.size(); ++i)
			{
				if(_fanout[i]==0) return true;
				if(_fanout[i]->getTypeStr()!="PO")
					if(_fanout[i]->isfoutfl()) return true;
			}*/
		return false;
	}
	virtual bool isfin1Inv() { return isInv(_fanin1); }
	virtual bool isfin2Inv() { return isInv(_fanin2); }
	virtual bool whichfin(CirGate* comp)
	{
		if(isInvRe(_fanin1) == comp) return 0;
		else return 1;
	}
//sweep
	virtual void sweepaig()
	{
		if(_fanin1 != 0)
			isInvRe(_fanin1)->erasefout(this);
		if(_fanin2 != 0)
			isInvRe(_fanin2)->erasefout(this);
		for(size_t i=0; i<_fanout.size(); ++i)
			_fanout[i]->erasefin(this);
	}
	virtual void erasefin(CirGate* fin)
	{
		if(_fanin1 != 0)
			if(isInvRe(_fanin1) == fin)
				_fanin1 = 0;
		if(_fanin2 != 0)
			if(isInvRe(_fanin2) == fin)
				_fanin2 = 0;
	}
//optimize
	virtual void optgate(CirGate*& rep, bool& inv)
	{
		for(size_t i=0; i<_fanout.size(); ++i)
		{
			if(_fanout[i]->getTypeStr() == "AIG")
			{
				if(!_fanout[i]->whichfin(this))
				{
					if(!_fanout[i]->isfin1Inv())
					{
						_fanout[i]->changefin1(rep);
						if(inv) _fanout[i]->setFin1Inv();
					}
					else
					{
						_fanout[i]->changefin1(rep);
						if(!inv) _fanout[i]->setFin1Inv();
					}
				}
				else
				{
					if(!_fanout[i]->isfin2Inv())
					{
						_fanout[i]->changefin2(rep);
						if(inv) _fanout[i]->setFin2Inv();
					}
					else
					{
						_fanout[i]->changefin2(rep);
						if(!inv) _fanout[i]->setFin2Inv();
					}
				}
			}
			else if(_fanout[i]->getTypeStr() == "PO")
			{
				if(!_fanout[i]->isfinInv())
				{
					_fanout[i]->changefin(rep);
					if(inv) _fanout[i]->setFinInv();
				}
				else
				{
					_fanout[i]->changefin(rep);
					if(!inv) _fanout[i]->setFinInv();
				}
			}
			isInvRe(rep)->setFanout(_fanout[i]);
		}
		isInvRe(_fanin1)->erasefout(this);
		isInvRe(_fanin2)->erasefout(this);
	}
	virtual void changefin1(CirGate*& fin1) { _fanin1 = fin1; }
	virtual void changefin2(CirGate*& fin2) { _fanin2 = fin2; }
//simulate
	virtual void simulate()
	{
		if(_fanin1!=0 && _fanin2!=0)
		{
			if(isfin1Inv() && isfin2Inv())
				{ _sim = ~(isInvRe(_fanin1)->getsim()) & ~(isInvRe(_fanin2)->getsim()); }
			else if(isfin1Inv() && !isfin2Inv())
				{ _sim = ~(isInvRe(_fanin1)->getsim()) & _fanin2->getsim(); }
			else if(!isfin1Inv() && isfin2Inv())
				{ _sim = _fanin1->getsim() & ~(isInvRe(_fanin2)->getsim()); }
			else
				{ _sim = _fanin1->getsim() & _fanin2->getsim(); }
		}
	}
//DFS
	virtual void DFS(vector<CirGate*>& _list)
	{//cout<<"aig "<<ID<<endl;
		if(!isfin1fl())
		{
			if(!isInvRe(_fanin1)->ismarked())
				isInvRe(_fanin1)->DFS(_list);
		}
		if(!isfin2fl())
		{
			if(!isInvRe(_fanin2)->ismarked())
				isInvRe(_fanin2)->DFS(_list);
		}
		setmark();
		_list.push_back(this);
	}

private:
	CirGate* _fanin1;
	CirGate* _fanin2;
};

class Const0 : public CirGate
{
public:
	Const0(int i=0, int j=0) : CirGate(i, j), _fanin(0) {}
	virtual ~Const0() {}

   string getTypeStr() const { return string("CONST"); }
	void printGate() const
	{
		string temp = "= ";
		string a = static_cast<ostringstream*>(&(ostringstream()<<ID))->str();
		temp.append("CONST(").append(a).append(")");
		a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
		temp.append(", line ").append(a);
		cout<<setw(49)<<left<<temp;
		cout<<right<<"="<<endl;
		cout<<"= FECs: ";
		if(_grpptr == 0)
			cout<<"                                         ="<<endl;
		else
		{
			vector<CirGate*> list = *_grpptr;
			sort(list.begin(), list.end(), CompGate());
			temp = "";
			for(size_t i=0; i<list.size(); ++i)
			{
				a = static_cast<ostringstream*>(&(ostringstream()<<(list[i]->getID())))->str();
				if(list[i]->getsim() == _sim && list[i]->getID() != ID)
					temp.append(a).append(" ");
				else if(list[i]->getID() != ID)
					temp.append("!").append(a).append(" ");
			}
			cout<<setw(41)<<left<<temp;
			cout<<right<<"="<<endl;
		}
	}
	void reportFanin(int level, int count=0, bool inv=0) const
	{
		assert(level >= 0);
		for(int i=0; i<2*count; ++i) cout<<" ";
		if(inv) cout<<"!";
		cout<<"CONST 0"<<endl;
	}
//DFS
	virtual void DFS(vector<CirGate*>& _list)
	{
		setmark();
		_list.push_back(this);
	}

private:
	CirGate* _fanin;
};
class UNDEF : public CirGate
{
public:
	UNDEF(int i=0, int j=0) : CirGate(i, j) {}
	virtual ~UNDEF() {}

   string getTypeStr() const { return string("UNDEF"); }
	void printGate() const
	{
		string temp = "= ";
		string a = static_cast<ostringstream*>(&(ostringstream()<<ID))->str();
		temp.append("UNDEF(").append(a).append(")");
		a = static_cast<ostringstream*>(&(ostringstream()<<_line))->str();
		temp.append(", line ").append(a);
		cout<<setw(49)<<left<<temp;
		cout<<right<<"="<<endl;;
		cout<<"= FECs:                                          ="<<endl;
	}
	void reportFanin(int level, int count=0, bool inv=0) const
	{
		assert(level >= 0);
		for(int i=0; i<2*count; ++i) cout<<" ";
		if(inv) cout<<"!";
		cout<<"UNDEF "<<ID<<endl;
	}
//sweep
	virtual void sweepaig()
	{
		for(size_t i=0; i<_fanout.size(); ++i)
		{
			_fanout[i]->erasefin(this);
		}
	}
//DFS
	virtual void DFS(vector<CirGate*>& _list)	{ setmark(); }
};
#endif // CIR_GATE_H
