/****************************************************************************
  FileName     [ cirMgr.cpp ]
  PackageName  [ cir ]
  Synopsis     [ Define cir manager functions ]
  Author       [ Chung-Yang (Ric) Huang ]
  Copyright    [ Copyleft(c) 2008-2012 LaDs(III), GIEE, NTU, Taiwan ]
****************************************************************************/

#include <iostream>
#include <iomanip>
#include <cstdio>
#include <ctype.h>
#include <cassert>
#include <cstring>
#include <algorithm>
#include "cirMgr.h"
#include "cirGate.h"
#include "util.h"

using namespace std;

// TODO: Implement memeber functions for class CirMgr

/*******************************/
/*   Global variable and enum  */
/*******************************/
CirMgr* cirMgr = 0;

enum CirParseError {
   EXTRA_SPACE,
   MISSING_SPACE,
   ILLEGAL_WSPACE,
   ILLEGAL_NUM,
   ILLEGAL_IDENTIFIER,
   ILLEGAL_SYMBOL_TYPE,
   ILLEGAL_SYMBOL_NAME,
   MISSING_NUM,
   MISSING_IDENTIFIER,
   MISSING_NEWLINE,
   MISSING_DEF,
   CANNOT_INVERTED,
   MAX_LIT_ID,
   REDEF_GATE,
   REDEF_SYMBOLIC_NAME,
   REDEF_CONST,
   NUM_TOO_SMALL,
   NUM_TOO_BIG,

   DUMMY_END
};

/**************************************/
/*   Static varaibles and functions   */
/**************************************/
static unsigned lineNo = 0;  // in printint, lineNo needs to ++
static unsigned colNo  = 0;  // in printing, colNo needs to ++
//static char buf[1024];
static string errMsg;
static int errInt;
static CirGate *errGate;

static bool
parseError(CirParseError err)
{
   switch (err) {
      case EXTRA_SPACE:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Extra space character is detected!!" << endl;
         break;
      case MISSING_SPACE:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Missing space character!!" << endl;
         break;
      case ILLEGAL_WSPACE: // for non-space white space character
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Illegal white space char(" << errInt
              << ") is detected!!" << endl;
         break;
      case ILLEGAL_NUM:
         cerr << "[ERROR] Line " << lineNo+1 << ": Illegal "
              << errMsg << "!!" << endl;
         break;
      case ILLEGAL_IDENTIFIER:
         cerr << "[ERROR] Line " << lineNo+1 << ": Illegal identifier \""
              << errMsg << "\"!!" << endl;
         break;
      case ILLEGAL_SYMBOL_TYPE:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Illegal symbol type (" << errMsg << ")!!" << endl;
         break;
      case ILLEGAL_SYMBOL_NAME:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Symbolic name contains un-printable char(" << errInt
              << ")!!" << endl;
         break;
      case MISSING_NUM:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Missing " << errMsg << "!!" << endl;
         break;
      case MISSING_IDENTIFIER:
         cerr << "[ERROR] Line " << lineNo+1 << ": Missing \""
              << errMsg << "\"!!" << endl;
         break;
      case MISSING_NEWLINE:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": A new line is expected here!!" << endl;
         break;
      case MISSING_DEF:
         cerr << "[ERROR] Line " << lineNo+1 << ": Missing " << errMsg
              << " definition!!" << endl;
         break;
      case CANNOT_INVERTED:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": " << errMsg << " " << errInt << "(" << errInt/2
              << ") cannot be inverted!!" << endl;
         break;
      case MAX_LIT_ID:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Literal \"" << errInt << "\" exceeds maximum valid ID!!"
              << endl;
         break;
      case REDEF_GATE:
         cerr << "[ERROR] Line " << lineNo+1 << ": Literal \"" << errInt
              << "\" is redefined, previously defined as "
              << errGate->getTypeStr() << " in line " << errGate->getLineNo()
              << "!!" << endl;
         break;
      case REDEF_SYMBOLIC_NAME:
         cerr << "[ERROR] Line " << lineNo+1 << ": Symbolic name for \""
              << errMsg << errInt << "\" is redefined!!" << endl;
         break;
      case REDEF_CONST:
         cerr << "[ERROR] Line " << lineNo+1 << ", Col " << colNo+1
              << ": Cannot redefine const (" << errInt << ")!!" << endl;
         break;
      case NUM_TOO_SMALL:
         cerr << "[ERROR] Line " << lineNo+1 << ": " << errMsg
              << " is too small (" << errInt << ")!!" << endl;
         break;
      case NUM_TOO_BIG:
         cerr << "[ERROR] Line " << lineNo+1 << ": " << errMsg
              << " is too big (" << errInt << ")!!" << endl;
         break;
      default: break;
   }
   return false;
}

/**************************************************************/
/*   class CirMgr member functions for circuit construction   */
/**************************************************************/
bool
CirMgr::readCircuit(const string& fileName)
{//open the file
	ifstream* cir = new ifstream(fileName.c_str());
	vector<string> li;
	string str;
	if(!cir->is_open())
	{
		cout<<"Cannot open design \""<<fileName<<"\"!!"<<endl;
		cir->close();
		delete cir;
		return false;
	}
	while(!cir->eof())
	{
		getline(*cir, str);
		li.push_back(str);
	}
//parse
	string tok;
	size_t s_tok = 0;
	for(size_t i=0; i<6; ++i)
	{
		myStrGetTok(li[0], tok, s_tok);
		if(i == 1) 
		{
			myStr2Int(tok, M);
		}
		else if(i == 2) 
		{
			myStr2Int(tok, I);
		}
		else if(i == 3)
		{
			myStr2Int(tok, L);
		}
		else if(i == 4)
		{
			myStr2Int(tok, O);
		}
		else if(i == 5)
		{
			myStr2Int(tok, A);
		}
		s_tok += (tok.size()+1);
	}
//cout<<M<<" "<<I<<" "<<L<<" "<<O<<" "<<tok<<endl;
	
	for(int i=0; i<M+O+1; ++i)
			_inorder.push_back(0);

	size_t l = 0;
	for(l=1; l<li.size(); ++l)
		{ if(!isdigit(li[l].c_str()[0])) break; }
	size_t sym = l;

	//CONST0
	_inorder[0] = new Const0;
	//PI
	int id_pi;
	for(int i=1; i<(I+1); ++i )
	{
		if(myStr2Int(li[i], id_pi))
		{
			CirGate* a = new Pi(id_pi/2, i+1);
			_piList.push_back(a);
			_inorder[id_pi/2] = a;
			//cout!!!!!!!!!!!!!!!!!!!
			//cout<<id_pi<<endl;
		}
	}

	//AIG
	int id_aig;
	int aig[l-I-O-1][3];
	string aigtok;
	for(size_t i=I+O+1; i<l; ++i)
	{
		myStrGetTok(li[i], aigtok);
		if(myStr2Int(aigtok, id_aig))
		{
			aig[i-I-O-1][0] = id_aig;
			CirGate* a = new And(id_aig/2, i+1);
			//_aigList.push_back(a);
			_inorder[id_aig/2] = a;
			size_t ts = aigtok.size();
			for(int j=1; j<3; ++j)
			{
				myStrGetTok(li[i], aigtok, ts+1);
				myStr2Int(aigtok, aig[i-I-O-1][j]);
				ts += aigtok.size()+1;
			}
		}
	}
	for(size_t i=0; i<l-I-O-1; ++i)
	{
		if(_inorder[aig[i][1]/2] == 0)
			_inorder[aig[i][1]/2] = new UNDEF(aig[i][1]/2);
		if(_inorder[aig[i][2]/2] == 0)
			_inorder[aig[i][2]/2] = new UNDEF(aig[i][2]/2);

		_inorder[aig[i][0]/2]->setFanin(_inorder[aig[i][1]/2], _inorder[aig[i][2]/2]);
		if(aig[i][1]%2)
			{ _inorder[aig[i][0]/2]->setFin1Inv(); }
		if(aig[i][2]%2)
			{ _inorder[aig[i][0]/2]->setFin2Inv(); }
	}

//cout!!!!!!!!!!!!!!!!
/*for(size_t i=0; i<l-I-O-1; ++i)
{
	for(int j=0; j<3; ++j)
		cout<<aig[i][j]<<" ";
	cout<<endl;
}*/
	//PO
	int po_id;
	for(int i=M+1; i<M+O+1; ++i)
	{
		if(myStr2Int(li[I+i-M], po_id))
		{
			CirGate* a = new Po(i, I+1+i-M);
			if(_inorder[po_id/2] == 0)
				_inorder[po_id/2] = new UNDEF(po_id/2);

			a->setFanin(_inorder[po_id/2]);
			if(po_id%2) a->setFinInv();
			_poList.push_back(a);
			_inorder[i] = a;
			//cout!!!!!!!!!!!!!!!!!!!!!!!!
			//cout<<po_id<<endl;
		}
	}

	//symbol
	int pnum;
	string symtok;
	for( ; sym<li.size()-1; ++sym)
	{
		if(li[sym] == "c") break;
		if(li[sym].c_str()[0] == 'i')
		{
			if(myStrGetTok(li[sym], symtok, 1))
			{
				if(myStr2Int(symtok, pnum))
				{
					myStrGetTok(li[sym], symtok, symtok.size()+1);
					_piList[pnum]->setsym(symtok);
				}
			}
		}
		else if(li[sym].c_str()[0] == 'o')
		{
			if(myStrGetTok(li[sym], symtok, 1))
			{
				if(myStr2Int(symtok, pnum))
				{
					myStrGetTok(li[sym], symtok, symtok.size()+1);
					_poList[pnum]->setsym(symtok);
				}
			}
		}
		else {}
	}

	//floating
	/*for(int i=1; i<=M+O; ++i)
	{
		if(_inorder[i] != 0 && _inorder[i]->getTypeStr() != "UNDEF")
		{
			if(_inorder[i]->getTypeStr()=="PI" && _inorder[i]->isfoutfl())
				_floating.push_back(_inorder[i]);
			else if(_inorder[i]->getTypeStr()=="PO" && _inorder[i]->isfinfl())
				_floating.push_back(_inorder[i]);
			else if(_inorder[i]->getTypeStr()=="AIG" && (_inorder[i]->isfinfl() || _inorder[i]->isfoutfl()))
				_floating.push_back(_inorder[i]);
		}
	}*/
	makeFList();
	//DFS
	DFSList();

//cout!!!!!!!!!!!!!
/*for(int i=0; i<M+O+1; ++i)
{
	if(_inorder[i] != 0)
		cout<<i<<" "<<_inorder[i]->getID()<<endl;
}*/

	cir->close();
	delete cir;
	return true;
}

/**********************************************************/
/*   class CirMgr member functions for circuit printing   */
/**********************************************************/
/*********************
Circuit Statistics
==================
  PI          20
  PO          12
  AIG        130
------------------
  Total      167
*********************/
void
CirMgr::printSummary() const
{
	/*int A = 0;
	for(size_t i=0; i<_inorder.size(); ++i)
	{
		if(_inorder[i] != 0)
			if(_inorder[i]->getTypeStr() == "AIG")
				++A;
	}*/
	cout<<endl;
	cout<<"Circuit Statistics"<<endl;
	cout<<"=================="<<endl;
	cout<<"  PI"<<setw(12)<<_piList.size()<<endl;
	cout<<"  PO"<<setw(12)<<_poList.size()<<endl;
	cout<<"  AIG"<<setw(11)<<A<<endl;
	cout<<"------------------"<<endl;
	cout<<"  Total"<<setw(9)<<_piList.size()+_poList.size()+A<<endl;
}

void
CirMgr::printNetlist() const
{
	int id;
	cout<<endl;
	for(size_t i=0; i<_netList.size(); ++i)
	{
		cout<<"["<<i<<"] ";
		if(_netList[i]->getTypeStr() == "PI")
		{
			cout<<"PI  "<<_netList[i]->getID();
			if(_netList[i]->getsym() != "")
				cout<<" ("<<_netList[i]->getsym()<<")";
			cout<<endl;
		}
		if(_netList[i]->getTypeStr() == "PO")
		{
			cout<<"PO  "<<_netList[i]->getID()<<" ";
			if(_netList[i]->getfinID(id))
			{
				if(_netList[i]->isfinfl()) cout<<"*";
				cout<<id;
			}
			else
			{
				if(_netList[i]->isfinfl()) cout<<"*";
				cout<<"!"<<id;
			}
			if(_netList[i]->getsym() != "")
				cout<<" ("<<_netList[i]->getsym()<<")";
			cout<<endl;
		}
		if(_netList[i]->getTypeStr() == "AIG")
		{
			cout<<"AIG "<<_netList[i]->getID()<<" ";
			if(_netList[i]->getfinID(id))
			{
				if(_netList[i]->isfin1fl()) cout<<"*";
				cout<<id<<" ";
			}
			else
			{
				if(_netList[i]->isfin1fl()) cout<<"*";
				cout<<"!"<<id<<" ";
			}
			if(_netList[i]->getfinID2(id))
			{
				if(_netList[i]->isfin2fl()) cout<<"*";
				cout<<id;
			}
			else
			{
				if(_netList[i]->isfin2fl()) cout<<"*";
				cout<<"!"<<id;
			}
			cout<<endl;
		}
		if(_netList[i]->getTypeStr() == "CONST")
		{ cout<<"CONST0"<<endl; }
	}
}

void
CirMgr::printPIs() const
{
   cout << "PIs of the circuit:";
   for(size_t i=0; i<_piList.size(); ++i)
		cout<<" "<<_piList[i]->getID();
	cout << endl;
}

void
CirMgr::printPOs() const
{
   cout << "POs of the circuit:";
   for(size_t i=0; i<_poList.size(); ++i)
		cout<<" "<<_poList[i]->getID();
   cout << endl;
}

void
CirMgr::printFloatGates() const
{
	if(_floating.size()!=0)
	{
		vector<size_t> _ffin;
		vector<size_t> _ffout;
		for(size_t i=0; i<_floating.size(); ++i)
		{
			if(_floating[i]->getTypeStr() == "PI")
				_ffout.push_back(_floating[i]->getID());
			if(_floating[i]->getTypeStr() == "PO")
				_ffin.push_back(_floating[i]->getID());
			if(_floating[i]->getTypeStr() == "AIG")
			{
				if(_floating[i]->isfoutfl())
					_ffout.push_back(_floating[i]->getID());
				if(_floating[i]->isfinfl())
					_ffin.push_back(_floating[i]->getID());
			}
		}
		if(_ffin.size()>0)
		{
			cout<<"Gates with floating fanin(s):";
			for(size_t i=0; i<_ffin.size(); ++i)
				cout<<" "<<_ffin[i];
			cout<<endl;
		}
		if(_ffout.size()>0)
		{
			cout<<"Gates defined but not used  :";
			for(size_t i=0; i<_ffout.size(); ++i)
				cout<<" "<<_ffout[i];
		}
		cout<<endl;
	}
}

void
CirMgr::printFECPairs() const
{
	int count = 0;
	vector<CirGate*> list;
	for(size_t i=0; i<_fecList.size(); ++i)
	{
		list = *_fecList[i];
		sort(list.begin(), list.end(), CompGate());
		if(list.size() > 1)
		{
			cout<<"["<<count<<"]";
			for(size_t j=0; j<list.size(); ++j)
			{
				cout<<" ";
				if(list[0]->getsim() != list[j]->getsim())
					cout<<"!";
				cout<<list[j]->getID();
			}
			cout<<endl;
			++count;
		}
	}
}
/*
bool
CirMgr::checkFECinv(CirGate* a, CirGate* b) const
{ return (a->getsim() != b->getsim()); }
*/
void
CirMgr::writeAag(ostream& outfile) const
{
	/*int A = 0;
	for(size_t i=0; i<_netList.size(); ++i)
	{
		if(_netList[i]->getTypeStr() == "AIG")
			++A;
	}*/
	outfile << "aag " << M << " " << I<< " " << L << " " << O << " " << A;
	
	for(size_t i=0; i<_piList.size(); ++i)
		outfile << "\n" << (_piList[i]->getID()*2);

	int id_1, id_2;
	for(size_t i=0; i<_poList.size(); ++i)
	{
		if(_poList[i]->getfinID(id_1)) 
			outfile << "\n" << (id_1*2);
		else
			outfile << "\n" << (id_1*2+1);
	}

	for(size_t i=0; i<_netList.size(); ++i)
	{
		if(_netList[i]->getTypeStr() == "AIG")
		{
			outfile << "\n" << (_netList[i]->getID()*2) << " ";
			if(_netList[i]->getfinID(id_1))
				outfile << (id_1*2) << " ";
			else
				outfile << (id_1*2+1) << " ";

			if(_netList[i]->getfinID2(id_2))
				outfile << (id_2*2);
			else
				outfile << (id_2*2+1);
		}
	}

	for(size_t i=0; i<_piList.size(); ++i)
	{
		if(_piList[i]->getsym() != "")
			outfile << "\n" << "i" << i << " " << _piList[i]->getsym();
	}

	for(size_t i=0; i<_poList.size(); ++i)
	{
		if(_poList[i]->getsym() != "")
			outfile << "\n" << "o" << i << " " << _poList[i]->getsym();
	}

	cout << "\n";

}

bool
CirMgr::checkfList(CirGate* cirg) const
{
	for(size_t i=0; i<_floating.size(); ++i)
		{ if(cirg == _floating[i]) return true; }
	return false;
}

void
CirMgr::makeFList()
{
	clearList(_floating);
	for(int i=1; i<=M+O; ++i)
	{
		if(_inorder[i] != 0 && _inorder[i]->getTypeStr() != "UNDEF")
		{
			if(_inorder[i]->getTypeStr()=="PI" && _inorder[i]->isfoutfl())
				_floating.push_back(_inorder[i]);
			else if(_inorder[i]->getTypeStr()=="PO" && _inorder[i]->isfinfl())
				_floating.push_back(_inorder[i]);
			else if(_inorder[i]->getTypeStr()=="AIG" && (_inorder[i]->isfinfl() || _inorder[i]->isfoutfl()))
				_floating.push_back(_inorder[i]);
		}
	}
}
void
CirMgr::DFSList()
{
	clearList(_netList);
	CirGate::setGlobal();
	for(size_t i=0; i<_poList.size(); ++i)
	{ _poList[i]->DFS(_netList); }
}
